/**********************************************************************
 * Copyright (c) 2007 IBM Corporation and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 * IBM - Initial API and implementation
 **********************************************************************/
package org.eclipse.cosmos.rm.smlif.internal.common;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.cosmos.rm.smlif.internal.SMLPlugin;
import org.eclipse.cosmos.rm.validation.internal.common.ISMLConstants;
import org.eclipse.cosmos.rm.validation.internal.common.IValidationConstants;
import org.eclipse.jface.dialogs.ErrorDialog;
import org.eclipse.osgi.util.NLS;
import org.eclipse.swt.SWT;
import org.eclipse.swt.dnd.Clipboard;
import org.eclipse.swt.dnd.TextTransfer;
import org.eclipse.swt.dnd.Transfer;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.MenuItem;
import org.eclipse.swt.widgets.Shell;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Provides common utility methods that used by multiple classes
 * under different packages.
 * 
 * @author Ali Mehregani
 */
public class SMLCommonUtil
{	
	/**
	 * Used when serializing a node into string
	 */
	private static boolean textNodeWrite = false;
	
	public static void openErrorWithDetail(Shell parent, String title, String message, String cause)
	{
		
		class CustomErrorDialog extends ErrorDialog
		{
			private String cause;
			private static final int LIST_ITEM_COUNT = 7;
			private org.eclipse.swt.widgets.List list;
			private boolean isListDisplayed;
			private Clipboard clipboard;
			
			public CustomErrorDialog(Shell parentShell, int severity, String dialogTitle, String message, String cause) 
			{
				super (parentShell, dialogTitle, message, new Status(severity, SMLPlugin.PLUGIN_ID, severity, "", new Throwable(cause)), IStatus.OK | IStatus.INFO | IStatus.WARNING | IStatus.ERROR);
				this.cause = cause;
				isListDisplayed = false;
			}
			
			public void openErrorDialog()
			{
				this.open();
			}			
			
		    			
		    protected org.eclipse.swt.widgets.List createDropDownList(Composite parent) 
		    {
		    	if (isListDisplayed)
		    	{
		    		isListDisplayed = false;		    		
		    		list.dispose();
		    		return list;
		    	}
		    	
		    	isListDisplayed = true;
		    	
		    	// create the list
		    	list = new org.eclipse.swt.widgets.List(parent, SWT.BORDER | SWT.H_SCROLL | SWT.V_SCROLL | SWT.MULTI);

		    	// fill the list
		    	if (cause != null)
		    	{
			    	StringTokenizer st = new StringTokenizer (cause, SMLConstants.nl);
			    	String indents = "";
			    	boolean addedIndentation = false;
			    	while (st.hasMoreTokens())
			    	{
			    		String currentToken = st.nextToken();
			    		currentToken = currentToken.replace ('\r', ' ');
			    		currentToken = currentToken.trim();
			    		if (!addedIndentation && currentToken.startsWith("at"))
			    		{
			    			indents += "  ";
			    			addedIndentation = true;
			    		}
			    		else if (addedIndentation && !currentToken.startsWith("at"))
			    			addedIndentation = false;
			    		
			    		
			    		list.add(indents + currentToken);
			    	}
		    	}
		    	
		        GridData data = new GridData(GridData.HORIZONTAL_ALIGN_FILL
		                | GridData.GRAB_HORIZONTAL | GridData.VERTICAL_ALIGN_FILL
		                | GridData.GRAB_VERTICAL);
		        data.heightHint = list.getItemHeight() * LIST_ITEM_COUNT;
		        data.horizontalSpan = 2;
		        list.setLayoutData(data);
		        list.setFont(parent.getFont());
		        
		        Menu copyMenu = new Menu(list);
		        MenuItem copyItem = new MenuItem(copyMenu, SWT.NONE);
		        copyItem.addSelectionListener(new SelectionListener() {
		            /*
		             * @see SelectionListener.widgetSelected (SelectionEvent)
		             */
		            public void widgetSelected(SelectionEvent e) {
		                copyToClipboard();
		            }

		            /*
		             * @see SelectionListener.widgetDefaultSelected(SelectionEvent)
		             */
		            public void widgetDefaultSelected(SelectionEvent e) {
		                copyToClipboard();
		            }
		        });
		        
		        copyItem.setText(SMLMessages.commonCopy);
		        list.setMenu(copyMenu);
		    	
		        return list;
		    }
		    
		    private void copyToClipboard() {
		        if (clipboard != null)
		            clipboard.dispose();
		        StringBuffer clipBoardBuffer = new StringBuffer();		       
		        
		        String[] listEntrySelections = list.getSelection();
		        for (int i = 0; i < listEntrySelections.length; i++)
		        	clipBoardBuffer.append(listEntrySelections[i] + SMLConstants.LINE_SEPARATOR);
		        clipboard = new Clipboard(list.getDisplay());
		        clipboard.setContents(new Object[] { clipBoardBuffer.toString() },
		                new Transfer[] { TextTransfer.getInstance() });
		    }


		}
				
		CustomErrorDialog errorDialog = new CustomErrorDialog (parent, IStatus.ERROR, title, message, cause);
		errorDialog.openErrorDialog();
	}
	
	public static void openErrorWithDetail(final String title, final String message, Throwable t)
	{		
		String stackTrace =getExceptionStackTrace(t);
		Throwable exceptionCause = (t.getCause() != null ? t.getCause() : t);
			
		final String cause = exceptionCause.getClass().getName() + SMLConstants.nl + exceptionCause.getMessage() + SMLConstants.nl + stackTrace;
		
		SMLPlugin.getDefault().getWorkbench().getDisplay().syncExec(new Runnable()
		{

			public void run() 
			{
				openErrorWithDetail(SMLPlugin.getDefault().getWorkbench().getActiveWorkbenchWindow().getShell(), title, message, cause);
				
			}});
		
	}
	
	public static String getExceptionStackTrace(Throwable t)
	{
		Throwable throwable = findCause(t);
				
		StringWriter stackTrace = new StringWriter(); 
		throwable.printStackTrace(new PrintWriter(stackTrace));	
		
		return stackTrace.toString();
	}
	
	public static Throwable findCause (Throwable t)
	{
		Throwable cause = null;
		if (t instanceof InvocationTargetException)
			cause = ((InvocationTargetException)t).getTargetException();			
		else if (t instanceof CoreException)
			cause = ((CoreException)t).getStatus().getException();				
		
		if (cause == null)
			return t;
		return findCause(cause);
	}
	
	
	public static int identifyDocumentType (IFile file)
	{
		Exception exception = null;
		try
		{
			return org.eclipse.cosmos.rm.validation.internal.common.SMLCommonUtil.identifyDocumentType(file);
		} 
		catch (IOException e)
		{
			exception = e;
		} 
		catch (CoreException e)
		{
			exception = e;
		}
		
		if (exception != null)
		{
			SMLCommonUtil.openErrorWithDetail(SMLMessages.errorContentTypeTitle, 
				NLS.bind(SMLMessages.errorContentType, file.getFullPath()), exception);
		}
		
		return ISMLConstants.TYPE_UNKNOWN;
	}
	
	
	/**
	 * Retrive the child node of 'parent' with the uri and localName passed in
	 * 
	 * @param parent The parent node
	 * @param uri The URI
	 * @param localName The local name of the element
	 * @return The element node or null if it cannot be found
	 */
	public static Node retrieveChildNode(Node parent, String uri, String localName)
	{
		if (parent == null)
			return null;
		
		NodeList children = parent.getChildNodes();		
		for (int i = 0, childCount = children.getLength(); i < childCount; i++)
		{
			Node child = children.item(i);
			if (uri.equals(child.getNamespaceURI()) && localName.equals(child.getLocalName()))
			{
				return child;				
			}
		}
		
		return null;
	}
	
	
	/**
	 * Retrieve the text child node of the argument node passed
	 * 
	 * @param node The node
	 * @return The child text of node
	 */
	public static String retrieveTextNode(Node node)
	{
		if (node == null)
			return null;
		
		StringBuffer sb = new StringBuffer();
		NodeList children = node.getChildNodes();
		for (int i = 0, childCount = children.getLength(); i < childCount; i++)
		{
			Node textNode = children.item(i);
			String currentText = null;
			if (textNode.getNodeType() == Node.TEXT_NODE && (currentText = textNode.getNodeValue()) != null
					&& (currentText = currentText.trim()).length() > 0)
			{
				sb.append(currentText);
			}
		}
		
		/* Walk through the node text and remove new lines and unnecessary white spaces */
		StringTokenizer st = new StringTokenizer(sb.toString(), SMLConstants.nl);
		StringBuffer finalTextValue = new StringBuffer();
		while (st.hasMoreTokens())
		{
			String currentLine = st.nextToken().trim();
			if (currentLine.length() <= 0)
				continue;
			
			if (finalTextValue.length() > 0)
				finalTextValue.append(" ");
			finalTextValue.append(currentLine);			
		}
		
		return finalTextValue.toString();
	}

	

	/**
	 * Equivalent to nestedNodeRetrieval (node, nodeNames, false)
	 */
	public static Node nestedNodeRetrieval(Node node, String[][] nodeNames)
	{
		return nestedNodeRetrieval (node, nodeNames, false);
	}
	
	
	/**
	 * Retrieve the nested node as specified by nodeNames (starting from 'node').
	 * The nodeNames are expected to be any array of String[] where String[0] = 
	 * URI of a node and String[1] = localName.
	 * 
	 * For example, given:
	 * <pre>
	 * 	<a><b><c></c></b></a>
	 * </pre>
	 * The node 'c' can be retrieved using
	 * nestedNodeRetrieval(node, new String[][] {new String[]{b's URI, "b"}, new String[]{c's URI, "c"}});
	 * where node corresponds to a's node.
	 * 	
	 * @param node The starting node
	 * @param nodeNames The node names
	 * @param create Creates the node if none is found
	 * @return The node representing the last entry of nodeNames or null if none 
	 * is found
	 */
	public static Node nestedNodeRetrieval(Node node, String[][] nodeNames, boolean create)
	{
		Node currentNode = node;
		for (int i = 0; i < nodeNames.length; i++)
		{
			if (currentNode == null)
				return null;
			
			Node childNode = retrieveChildNode(currentNode, nodeNames[i][0], nodeNames[i][1]);
			if (childNode == null && create)
			{
				childNode = currentNode.getOwnerDocument().createElementNS(nodeNames[i][0], nodeNames[i][1]);
				childNode.setPrefix(currentNode.getPrefix());
				currentNode.appendChild(childNode);				
			}
			currentNode = childNode;			
		}
		
		return currentNode;
		
	}
	
	
	/**
	 * Adds an XML element to the string buffer passed in.
	 * 
	 * @param stringBuffer The string buffer that the element will be written to
	 * @param indent The number of indents that should be prefixed to the element written
	 * @param elementName The element name
	 * @param close Indicates whether this element should be closed or not.  If set, 
	 * a forward slash is added before theelement is ended.  
	 * @param end Indicates if the element should be ended with a closed angle bracket.  
	 * For example elements requring attributes are not ended.
	 */
	public static void addElement (StringBuffer stringBuffer, int indent, String elementName, boolean close, boolean end)
	{
		if (indent > 0)
			addIndent(stringBuffer, indent);
		stringBuffer.append(SMLConstants.OPEN_ANGLE_BRACKET);
		stringBuffer.append(close ? SMLConstants.FORWARD_SLASH : IValidationConstants.EMPTY_STRING);		
		stringBuffer.append(elementName);
		stringBuffer.append(end ? SMLConstants.CLOSE_ANGLE_BRACKET + 
				(close ? IValidationConstants.LINE_SEPARATOR : IValidationConstants.EMPTY_STRING) : 
				IValidationConstants.EMPTY_STRING);		
	}

	
	/**
	 * Adds XML attributes to the string buffer passed in.  Attributes with a null value will
	 * not be written.
	 * 
	 * Pre-condition:
	 * <ul>
	 * 	<li> attributes and attributeValues must be valid (i.e. non-null) </li>
	 * 	<li> attributes and attributeValues length must be the same </li>
	 * </ul>
	 *  
	 * @param stringBuffer The string buffer that the attributes will be written to
	 * @param attributes attributes
	 * @param close If set, a forward slash is added after the attributes.  
	 * @param end If set, a closed angle bracket is added after the attributes
	 */
	public static void addAttribute(StringBuffer stringBuffer, NamedNodeMap attributes, boolean close, boolean end)
	{
		for (int i = 0, attCount = attributes.getLength(); i < attCount; i++)
		{
			Node currentAttribute = attributes.item(i);
			stringBuffer.append(SMLConstants.SINGLE_SPACE);
			stringBuffer.append(currentAttribute.getNodeName());
			stringBuffer.append(SMLConstants.EQUAL_SIGN);
			stringBuffer.append(SMLConstants.DOUBLE_QUOTE);
			stringBuffer.append(currentAttribute.getNodeValue());
			stringBuffer.append(SMLConstants.DOUBLE_QUOTE);
		}
		
		stringBuffer.append(close ? SMLConstants.FORWARD_SLASH : IValidationConstants.EMPTY_STRING);
		stringBuffer.append(end ? SMLConstants.CLOSE_ANGLE_BRACKET + (close ? IValidationConstants.LINE_SEPARATOR : IValidationConstants.EMPTY_STRING) : IValidationConstants.EMPTY_STRING);
	}
	
	public static void addIndent (StringBuffer stringBuffer, int indent)
	{
		for (int i = 0; i < indent; i++)
		{
			stringBuffer.append("\t");
		}
	}

	
	/**
	 * Retrieves and returns the child elements of 
	 * node (i.e. nodes of type Node.ELEMENT_NODE)
	 * 
	 * @param node The node
	 * @return Returns the child elements of node
	 */
	public static Node[] retrieveChildElements(Node node)
	{
		if (node == null)
			return new Node[0];
		
		List childElements = new ArrayList();
		NodeList children = node.getChildNodes();
		for (int i = 0, childCount = children.getLength(); i < childCount; i++)
		{
			Node child = children.item(i);
			if (child.getNodeType() == Node.ELEMENT_NODE)
				childElements.add(child);			
		}
		return (Node[])childElements.toArray(new Node[childElements.size()]);
	}
	
	
	public static Node createElement (Node parentNode, String uri, String localName)
	{
		Node node = parentNode.getOwnerDocument().createElementNS(uri, localName);
		node.setPrefix(parentNode.getPrefix());
		parentNode.appendChild(node);
		
		return node;
	}
	
	
		
	/**
	 * Equivalent to serializeNode(buffer, node, indent, null, null)
	 */
	public static void serialzeNode(StringBuffer buffer, Node node, int indent)
	{
		serialzeNode(buffer, node, indent, null, null);
	}
	
	
	
	/**
	 * Used to serialize a node into string
	 * 
	 * @param buffer The buffer that the string will be written to
	 * @param node The starting node
	 * @param indent The starting indent level (use 0 if starting at root)
	 * @param currentIndex The current index of the node (use null if not interested to store indices)
	 * @param indices The indices map (currentIndex must not be null if indices is used)
	 */
	public static void serialzeNode(StringBuffer buffer, Node node, int indent, String currentIndex, Map indices)
	{
		switch (node.getNodeType())
		{		
			case Node.TEXT_NODE:
			{
				textNodeWrite = true;
				buffer.append(node.getNodeValue().trim());
				break;				
			}
			case Node.COMMENT_NODE:
			{
				if (node.getNodeValue() == null)
					return;
				
				SMLCommonUtil.addIndent(buffer, indent);
				buffer.append("<!--" + node.getNodeValue() + "-->");
				break;
			}			
			case Node.ELEMENT_NODE:
			{
				boolean needsLineSep = false;
				for (int i = buffer.length() - 1; i >= 0; i--)
				{
					char currentChar = buffer.charAt(i);
					if (currentChar == '\n')
						break;
					else if (currentChar == '\t' || currentChar == ' ' || currentChar == '\r')
						continue;
					else
					{
						needsLineSep = true;
						break;
					}
				}
				
				if (needsLineSep)
					buffer.append(IValidationConstants.LINE_SEPARATOR);
				NamedNodeMap attributes = node.getAttributes();
				Node[] children = retrieveChildNodes(node);
				boolean hasAttributes = attributes != null && attributes.getLength() > 0;
				boolean hasChildren = children != null && children.length > 0;
				
				textNodeWrite = false;
				int[] elementIndices = new int[2];
				elementIndices[0] = buffer.length();		
				SMLCommonUtil.addElement(buffer, indent, node.getNodeName(), !hasAttributes && !hasChildren, !hasAttributes);
				if (hasAttributes)
					SMLCommonUtil.addAttribute(buffer, attributes, !hasChildren, true);
				if (!hasChildren)
				{
					elementIndices[1] = buffer.length();
					if (currentIndex != null)
						indices.put(currentIndex, elementIndices);
					return;
				}
				
				int counter = 0;
				for (int i = 0; i < children.length; i++)
				{
					Node child = children[i];
					if (child.getNodeType() == Node.TEXT_NODE)
					{
						String nodeValue = child.getNodeValue();
						if (nodeValue == null || nodeValue.trim().length() <= 0)
						{
							continue;
						}
					}
					else
					{
						buffer.append(IValidationConstants.LINE_SEPARATOR);
					}
						
					serialzeNode(buffer, child, indent + 1, currentIndex == null ? null : currentIndex + "," + counter++, indices); //$NON-NLS-1$
				}
						
						
				SMLCommonUtil.addElement(buffer, textNodeWrite ? -1 : indent, node.getNodeName(), true, true);
				elementIndices[1] = buffer.length();
				if (currentIndex != null)
					indices.put(currentIndex, elementIndices);
				textNodeWrite = false;
				break;
			}
			case Node.DOCUMENT_NODE:
			{
				Document document = (Document)node;
				boolean versionPresent = document.getXmlVersion() != null;
				boolean encodingPresent = document.getXmlEncoding() != null;
				
				if (versionPresent || encodingPresent)
				{
					buffer.append("<?xml" + (versionPresent ? " version=\"" + document.getXmlVersion() + "\"" : IValidationConstants.EMPTY_STRING) +
							(encodingPresent ? " encoding=\"" + document.getXmlEncoding() + "\"" : IValidationConstants.EMPTY_STRING) + "?>" + IValidationConstants.LINE_SEPARATOR);
				}
				NodeList children = node.getChildNodes();
				if (children == null)
					break;
				
				for (int i = 0, childCount = children.getLength(); i < childCount; i++)
				{
					serialzeNode(buffer, children.item(i), indent, currentIndex, indices);
				}
			}
		}
		
	}

	private static Node[] retrieveChildNodes(Node node)
	{
		if (node == null)
			return null;
		
		NodeList children = node.getChildNodes();	
		List finalChildrenList = new ArrayList();
		for (int i = 0, childCount = children.getLength(); i < childCount; i++)
		{
			Node child = children.item(i);
			String nodeValue = null;
			if (child.getNodeType() == Node.TEXT_NODE && ((nodeValue = child.getNodeValue()) == null || nodeValue.trim().length() <= 0))
			{
				continue;				
			}
			finalChildrenList.add(child);
		}
		
		return (Node[])finalChildrenList.toArray(new Node[finalChildrenList.size()]);
	}
}
