/**********************************************************************
 * Copyright (c) 2006 IBM Corporation and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 * IBM - Initial API and implementation
 **********************************************************************/
package org.eclipse.cosmos.rm.validation.internal.common;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.Hashtable;
import java.util.Map;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParserFactory;

import org.eclipse.cosmos.rm.validation.internal.artifacts.ElementTypeMap;
import org.eclipse.cosmos.rm.validation.internal.artifacts.TypeNode;
import org.eclipse.cosmos.rm.validation.internal.databuilders.AbstractDataBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.DataBuilderRegistry;
import org.eclipse.cosmos.rm.validation.internal.databuilders.ElementTypeMapDataBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.IdentityDataBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.RefTypeElementBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.TypeInheritanceDataBuilderImpl;
import org.eclipse.cosmos.rm.validation.internal.databuilders.IdentityDataBuilder.SMLIFIdentity;
import org.eclipse.hyades.test.common.junit.HyadesTestCase;
import org.w3c.dom.Document;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

/**
 * Contains common utility methods used by multiple test classes
 * 
 * @author Ali Mehregani
 */
public class UtilityTester
{
	/**
	 * Creates a dummy inheritance map based on the data that is passed in
	 * 
	 * @param nodes The relationship - an array of TypeNode[2], where
	 * TypeNode[0] = type TypeNode[1] = parent
	 */
	public static void registerInheritanceMap(TypeNode[][] nodes)
	{
		final Map inheritanceMap = new Hashtable();
		for (int i = 0; i < nodes.length; i++)
		{
			Map typePerUri = (Map)SMLCommonUtil.retrieveNestedMap(inheritanceMap, nodes[i][0].getUri(), true);
			typePerUri.put(nodes[i][0].getType(), nodes[i][1]);
		}		
		
		/* Build an inheritance map */
		DataBuilderRegistry.instance().registerDataStructureBuilder(TypeInheritanceDataBuilderImpl.ID, new TypeInheritanceDataBuilderImpl()
		{
			public Object getDataStructure()
			{			
				return inheritanceMap;
			}
		});
	}
	

	/**
	 * Parses the document passed in 
	 * 
	 * @param resource The resource to read from
	 * @param contentHandler The content handler
	 * 
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 * @throws IOException
	 */
	public static void parseDocument(InputStream resource, DefaultHandler contentHandler) throws ParserConfigurationException, SAXException, IOException
	{
		SAXParserFactory saxParserFactory = SAXParserFactory.newInstance();		
		saxParserFactory.setFeature("http://xml.org/sax/features/namespaces", true);		
		
		saxParserFactory.newSAXParser().parse(resource, contentHandler);			
	}
	
	
	/**
	 * Uses a DOM parser to parse the XML file passed in
	 * 
	 * @param resource The resource to be parsed
	 * @return The document representing the parsed resource
	 */
	public static Document domParseDocument(Object resource) throws ParserConfigurationException, SAXException, IOException 
	{
		DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.newInstance();
        docBuilderFactory.setNamespaceAware(true);
        docBuilderFactory.setIgnoringElementContentWhitespace(true);
        DocumentBuilder docBuilder = docBuilderFactory.newDocumentBuilder();
        
		return resource instanceof InputStream ? docBuilder.parse((InputStream)resource) : docBuilder.parse((InputSource)resource);
	}


	/**
	 * Registers an element type map with the data builder registery
	 * 
	 * @param elementTypeEntries
	 */
	public static void registerElementTypeMap(String[][] elementTypeEntries)
	{
		final ElementTypeMap elementTypeMap = new ElementTypeMap();
		for (int i = 0; i < elementTypeEntries.length; i++)
		{
			elementTypeMap.addElementDeclaration(elementTypeEntries[i][0], elementTypeEntries[i][1], elementTypeEntries[i][2]);
		}
		
		DataBuilderRegistry.instance().registerDataStructureBuilder(ElementTypeMapDataBuilder.ID, 
				new ElementTypeMapDataBuilder()
				{
					public Object getDataStructure()
					{
						return elementTypeMap;
					}	
				});
	}
	
	
	/**
	 * Constructs and registers an identity structure
	 * 
	 * @param name The name
	 * @param desc The description
	 * @param baseURI The base URI
	 */
	public static void registerIdentityStructure(String name, String desc, String baseURI)
	{
		final SMLIFIdentity smlIfIdentity = new SMLIFIdentity();
		smlIfIdentity.setName(name);
		smlIfIdentity.setDescription(desc);
		smlIfIdentity.setBaseURI(baseURI);
		
		DataBuilderRegistry.instance().registerDataStructureBuilder(IdentityDataBuilder.ID, 
				new IdentityDataBuilder()
				{
					public Object getDataStructure()
					{
						return smlIfIdentity;
					}	
				});
	}
	
	
	/**
	 * Compares two files passed in
	 * 
	 * @param expectedFile The expected file path
	 * @param actualFile The actual file path
	 * @throws IOException In case of an input/output error
	 */
	public static void compareFiles(String expectedFile, String actualFile) throws IOException
	{		
		BufferedReader expectedResults = new BufferedReader(new InputStreamReader(UtilityTester.class.getClassLoader().getResourceAsStream(expectedFile)));
		BufferedReader actualResults = new BufferedReader(new InputStreamReader(new FileInputStream(actualFile)));
		
		String currentExpectedLine = null;

		while ((currentExpectedLine = expectedResults.readLine()) != null)
		{
			HyadesTestCase.assertEquals("The content of the generated file is different than the expected file", currentExpectedLine, actualResults.readLine());
		}

		HyadesTestCase.assertNull("The actual file has more content than expected", actualResults.readLine());
		actualResults.close();
		
		File file = new File (actualFile);
		file.delete();
	}
	
	
	public static void registerReftypeElements (String[][] reftypes)
	{
		final Map reftypesMap = new Hashtable();
		for (int i = 0; i < reftypes.length; i++)
		{
			Map reftypesPerUri = SMLCommonUtil.retrieveNestedMap(reftypesMap, reftypes[i][0], true);
			reftypesPerUri.put(reftypes[i][1], IValidationConstants.EMPTY_STRING);			
		}
		
		DataBuilderRegistry.instance().registerDataStructureBuilder(RefTypeElementBuilder.ID, new AbstractDataBuilder(){

			public Object getDataStructure()
			{
				return reftypesMap;
			}
		});
	}
}
