/**********************************************************************
 * Copyright (c) 2006 IBM Corporation and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 * IBM - Initial API and implementation
 **********************************************************************/
package org.eclipse.cosmos.rm.validation.internal.common;

import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.Hashtable;
import java.util.Map;

import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathFactory;
import javax.xml.xpath.XPathFunction;
import javax.xml.xpath.XPathFunctionResolver;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.content.IContentType;
import org.eclipse.cosmos.rm.validation.internal.content.GenicContentDescriber;
import org.eclipse.cosmos.rm.validation.internal.content.PhenicContentDescriber;
import org.eclipse.cosmos.rm.validation.internal.content.SMLIFContentDescriber;
import org.eclipse.cosmos.rm.validation.internal.databuilders.DataBuilderRegistry;
import org.eclipse.cosmos.rm.validation.internal.databuilders.IDataBuilder;
import org.eclipse.cosmos.rm.validation.internal.reference.DerefXPathFunction;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;

/**
 * Provides common utility method used by multiple classes under
 * different pacakges 
 * 
 * @author Ali Mehregani
 */
public class SMLCommonUtil
{
	/**
	 * A static XPath instance used to compile XPath expressions
	 */
	public static XPath xpath;
	
	static
	{
		XPathFactory factory = XPathFactory.newInstance();
	    xpath = factory == null ? null : factory.newXPath();
	    if (xpath != null)
	    {
	    	/* Register the dref XPath function extension */
	    	xpath.setXPathFunctionResolver(new XPathFunctionResolver()
	    	{
	    		/**
	    		 * @see XPathFunctionResolver#resolveFunction(QName, int)
	    		 */
				public XPathFunction resolveFunction(QName functionName, int arity)
				{
					if (ISMLConstants.SML_FN_URI.equals(functionName.getNamespaceURI()) || ISMLConstants.SML_FN_URI2.equals(functionName.getNamespaceURI()) && 
						ISMLConstants.DEREF_FN.equals(functionName.getLocalPart()) &&
						arity == DerefXPathFunction.ARITY)
						return DerefXPathFunction.instance();
					return null;
				}
			});
	    }
	}
	
	/**
	 * A helper method provided to retrieve the data structure for a databuilder
	 * with the given id.
	 * 
	 * @param id The data builder id
	 */
	public static Object retrieveDataStructure(String databuilderId)
	{		
		IDataBuilder databuilder = DataBuilderRegistry.instance().getDataStructureBuilder(databuilderId);
		if (databuilder == null)
			return null;
		
		return databuilder.getDataStructure();
	}

	
	/**
	 * Removes line breaks and any spaces that follow a line break
	 * 
	 * @param str The string
	 * @return str with line break and white spaces that follow line breaks
	 * removed. 
	 */
	public static String removeLineBreaks(String str)
	{
		if (str == null)
			return null;
		
		StringBuffer sb = new StringBuffer();
		char currentChar, lastChar = 0;
		for (int i = 0, strLength = str.length(); i < strLength; i++)
		{
			currentChar = str.charAt(i);
				
			if (currentChar == '\r' || (lastChar == '\n' && (currentChar == '\t' || currentChar == ' ')))
			{
				continue;
			}
			
			if (currentChar != '\n')
			{
				sb.append(currentChar);				
			}
			lastChar = currentChar;
		}
		
		return sb.toString();
	}
	
	public static Map retrieveNestedMap (Map map, String key, boolean create)
	{
		Map nestedMap = (Map)map.get(key);
		if (create && nestedMap == null)
		{
			nestedMap = new Hashtable();
			map.put(key, nestedMap);
		}
		return nestedMap;
	}


	/**
	 * This method is used to retrieve an XML document referenced
	 * by a URI.
	 * 
	 * @param uri The uri where the document resides
	 * @return The root node of the XML document
	 */
	public static Node retrieveRemoteDocument(String uri) throws RemoteRetrievalException
	{
		Exception exception = null;
		try
		{
			URL locator = new URL(uri);
			InputStream stream = locator.openStream();			
			DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.newInstance();
	        docBuilderFactory.setNamespaceAware(true);
	        docBuilderFactory.setIgnoringElementContentWhitespace(true);
	        Document document = docBuilderFactory.newDocumentBuilder().parse((InputStream)stream);
	        return document.getChildNodes() == null || document.getChildNodes().getLength() != 1 ? null : document.getChildNodes().item(0);
		} 
		catch (IOException e)
		{
			exception = e;
		} 
		catch (SAXException e)
		{
			exception = e;
		} 
		catch (ParserConfigurationException e)
		{
			exception = e;
		}	
		
		throw new RemoteRetrievalException(exception);
	}
	
	
	/**
	 * Thrown when there is an error while retrieving a remote 
	 * document
	 * 
	 * @author Ali Mehregani
	 */
	public static class RemoteRetrievalException extends Exception
	{
		/**
		 * The serial version UID 
		 */
		private static final long serialVersionUID = 7633895444717500045L;
	
		public RemoteRetrievalException(String message)
		{
			super(message);
		}
		
		public RemoteRetrievalException(Exception cause)
		{
			super(cause);
		}
	}
	
	
	/**
	 * Uses a DOM parser to parse the XML file passed in
	 * 
	 * @param resource The resource to be parsed
	 * @return The document representing the parsed resource
	 */
	public static Document domParseDocument(InputStream resource, boolean ignoreComments) throws ParserConfigurationException, SAXException, IOException 
	{
		DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.newInstance();
        docBuilderFactory.setNamespaceAware(true);
        docBuilderFactory.setIgnoringElementContentWhitespace(true);
        docBuilderFactory.setIgnoringComments(ignoreComments);
        DocumentBuilder docBuilder = docBuilderFactory.newDocumentBuilder();
        
		return docBuilder.parse((InputStream)resource);
	}
	
	
	public static Document domParseDocument(InputStream resource) throws ParserConfigurationException, SAXException, IOException 
	{
		return domParseDocument(resource, true);
	}
	
	
	/**
	 * Returns {@link ISMLConstants#TYPE_GENIC} if the document type is determined to be genic,
	 * {@link ISMLConstants#TYPE_PHENIC} if phenic, {@link ISMLConstants#TYPE_SMLIF}
	 * if SML-IF, and {@link ISMLConstants#TYPE_UNKNOWN} if content type is unknown
	 * 
	 * @param file The file to be checked 
	 * @return The content type of the file passed in
	 * @throws CoreException 
	 * @throws IOException 
	 */
	public static int identifyDocumentType (IFile file) throws IOException, CoreException
	{
		InputStream fileInputStream = file.getContents();
		try
		{
			int type = ISMLConstants.TYPE_UNKNOWN;
	
			IContentType[] types = Platform.getContentTypeManager().findContentTypesFor(fileInputStream, file.getName());
			for (int i = 0; i < types.length; i++)
			{
				if (SMLIFContentDescriber.ID.equals(types[i].getId()))
				{
					return ISMLConstants.TYPE_SMLIF;
				}
				else if (GenicContentDescriber.ID.equals(types[i].getId()))
				{
					return ISMLConstants.TYPE_GENIC;
				}
				else if (PhenicContentDescriber.ID.equals(types[i].getId()))
				{
					type = ISMLConstants.TYPE_PHENIC;
				}
			}
			
			return type;
		}
		finally
		{
			if (fileInputStream != null)
				fileInputStream.close();
		}
	}
}
