/**********************************************************************
 * Copyright (c) 2006 IBM Corporation and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 * IBM - Initial API and implementation
 **********************************************************************/
package org.eclipse.cosmos.rm.validation.internal.databuilders;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.xml.namespace.NamespaceContext;

import org.eclipse.cosmos.rm.validation.internal.SMLActivator;
import org.eclipse.cosmos.rm.validation.internal.common.ISMLConstants;
import org.eclipse.cosmos.rm.validation.internal.common.IValidationConstants;
import org.eclipse.cosmos.rm.validation.internal.common.SMLCommonUtil;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

/**
 * This builder will construct a data structure that will be used
 * as part of the identity constraint (key, keyref, and unique) 
 * validation.  The data structure returned by this builder is of type
 * {@link IdentityConstraintStructure}
 * 
 * @author Ali Mehregani
 */
public class IdentityConstraintDataBuilder extends AbstractDataBuilder
{
	/**
	 * The ID of this builder
	 */
	public static final String ID = SMLActivator.PLUGIN_ID + ".IdentityConstraintDataBuilder";
	
	/**
	 * The identity constriant structure
	 */
	private IdentityConstraintStructure identityStruct;
	
	/**
	 * The current identity constraint declaration
	 */
	private IdentityConstraintDecl identityDecl;
	
	/**
	 * Indicates that the 'element' element is hit
	 */
	private boolean elementHit;
	
	/**
	 * Indicates that the 'annotation' element is hit
	 */
	private boolean annotationHit;
	
	/**
	 * Inidcates that the 'appinfo' element is hit
	 */
	private boolean appInfoHit;
	
	/**
	 * The current index of the document node
	 */
	private LinkedList documentNodeInx;
	
	/**
	 * Indicates that the 'document' element has been hit
	 */
	private boolean documentElementHit;
	
	/**
	 * Indicates that the 'data' element is hit
	 */
	private boolean dataElementHit;

	/**
	 * The current aliases of a document
	 */
	private List currentAliases;
	
	/**
	 * The current alias
	 */
	private String currentAlias;

	/**
	 * Indicates that an alias element has been hit
	 */
	private boolean aliasElementHit;
	
	/**
	 * The current index of the child element being processed
	 */
	private int currentInx;
	
	/**
	 * The current depth of the XML structure
	 */
	private int currentDepth;

	/**
	 * The uri of the element being processed
	 */
	private String currentURI;

	/**
	 * The local name of the element being processed
	 */
	private String currentElementName;
	
	/**
	 * The current orphan document index
	 */
	private int orphanInx;
	
	public IdentityConstraintDataBuilder()
	{
		identityStruct = new IdentityConstraintStructure();
		documentNodeInx = new LinkedList();
		currentDepth = -1;
		orphanInx = 0;
	}
	
	/**
	 * @see org.eclipse.cosmos.rm.validation.internal.databuilders.AbstractDataBuilder#startElement(java.lang.String, java.lang.String, java.lang.String, org.xml.sax.Attributes)
	 */
	public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException
	{		
		super.startElement(uri, localName, qName, attributes);
		
		byte currentPhase = getCurrentPhase();
		switch (currentPhase)
		{
			case ISMLConstants.GENICS_PHASE:
				handleGenicStartElement(uri, localName, attributes);							
				break;
			case ISMLConstants.PHENICS_PHASE:
				handlePhenicStartElement(uri, localName, attributes);
				break;
		}		
	}
	
	public void endElement(String uri, String localName, String qName) throws SAXException
	{
		super.endElement(uri, localName, qName);
		byte currentPhase = getCurrentPhase();
		switch (currentPhase)
		{
			case ISMLConstants.GENICS_PHASE:
				handleGenicCloseElement(uri, localName);							
				break;
			case ISMLConstants.PHENICS_PHASE:
				handlePhenicCloseElement(uri, localName);
				break;
		}		
	}

	private boolean validField(String field)
	{
		return field != null && field.length() > 0;
	}



	private void handleGenicStartElement(String uri, String localName, Attributes attributes)
	{
		if (ISMLConstants.SCHEMA_URI.equals(uri))
		{
			if (!elementHit && ISMLConstants.ELEMENT_ELEMENT.equals(localName))
			{
				elementHit = true;
				currentURI = getTargetNamespace();
				currentElementName = attributes.getValue(ISMLConstants.NAME_ATTRIBUTE);				
				currentElementName = currentElementName == null ? attributes.getValue(ISMLConstants.SCHEMA_URI, ISMLConstants.NAME_ATTRIBUTE) : currentElementName;
			}
			else if (elementHit && ISMLConstants.ANNOTATION_ELEMENT.equals(localName))
			{
				annotationHit = true;
			}
			else if (annotationHit && ISMLConstants.APP_INFO_ELEMENT.equals(localName))
			{
				appInfoHit = true;
			}			
			return;
		}
		
		if (appInfoHit && ISMLConstants.SML_URI.equals(uri))			
		{
			boolean isKey = ISMLConstants.KEY_ELEMENT.equals(localName);
			boolean isKeyRef = !isKey && ISMLConstants.KEY_REF_ELEMENT.equals(localName);
			boolean isUnique = !isKeyRef && !isKey && ISMLConstants.UNIQUE_ELEMENT.equals(localName);
			if (isKey || isKeyRef || isUnique)
			{
				 /* Get the name of the key */
				 String name = retrieveAttribute(attributes, ISMLConstants.SML_URI, ISMLConstants.NAME_ATTRIBUTE);
				 if (!validField(name))
					 return;
				 
				 identityDecl = new IdentityConstraintDecl(name);	
				 final Map clonedPrefixMap = (Map)((Hashtable)super.getPrefixMap()).clone();
				 
				 identityDecl.setNamespaceContext(new NamespaceContext(){

					public String getNamespaceURI(String prefix)
					{		
						String uri = (String)clonedPrefixMap.get(prefix);
						return uri == null ? IValidationConstants.EMPTY_STRING : uri;
					}

					public String getPrefix(String arg0)
					{
						/* Doesn't need to be implemented */
						return null;
					}

					public Iterator getPrefixes(String arg0)
					{
						/* Doesn't need to be implemented */
						return null;
					}});
				 
				 identityDecl.setType(isKey ? IdentityConstraintDecl.KEY_TYPE : (isKeyRef ? IdentityConstraintDecl.KEY_REF_TYPE : IdentityConstraintDecl.UNIQUE_TYPE));
				 if (isKeyRef)
				 {
					 identityDecl.setReference(retrieveAttribute(attributes, ISMLConstants.SML_URI, ISMLConstants.REFER_ATTRIBUTE));
				 }
			 }
			 else if (identityDecl != null)
			 {
				 if (ISMLConstants.SELECTOR_ELEMENT.equals(localName))
				 {					 
					 String selector = attributes.getValue(ISMLConstants.XPATH_ATTRIBUTE);
					 selector = selector == null ? attributes.getValue(ISMLConstants.SML_URI, ISMLConstants.XPATH_ATTRIBUTE) : selector;
					 identityDecl.setSelector(selector);
				 }
				 else if (ISMLConstants.FIELD_ELEMENT.equals(localName))
				 {
					 String field = attributes.getValue(ISMLConstants.XPATH_ATTRIBUTE);
					 field = field == null ? attributes.getValue(ISMLConstants.SML_URI, ISMLConstants.XPATH_ATTRIBUTE) : field;
					 identityDecl.addField(field);
				 }
			 }
		}		
	}

	
	private String retrieveAttribute(Attributes attributes, String uri, String localName)
	{
		String value = attributes.getValue(localName);
		return value == null ? attributes.getValue(uri, localName) : value;
	}

	private void handleGenicCloseElement(String uri, String localName)
	{
		if (ISMLConstants.SML_URI.equals(uri))
		{
			if (appInfoHit && identityDecl != null)
			{
				boolean isKey = ISMLConstants.KEY_ELEMENT.equals(localName);
				boolean isKeyref = ISMLConstants.KEY_REF_ELEMENT.equals(localName);
				boolean isUnique = ISMLConstants.UNIQUE_ELEMENT.equals(localName);
				
				if (isKey || isKeyref || isUnique)
				{
					/* The selector and the field are required parts of an identity constraint declaration */
					if (validField(identityDecl.getSelector()) && identityDecl.getFields().size() > 0)
					{
						identityStruct.addDeclaration(currentURI, currentElementName, identityDecl);
					}
					identityDecl = null;
				}
			}
			else if (appInfoHit && ISMLConstants.APP_INFO_ELEMENT.equals(localName))
			{
				appInfoHit = false;
			}
			else if (annotationHit && ISMLConstants.ANNOTATION_ELEMENT.equals(localName))
			{
				annotationHit = false;
			}
		}
		else if (ISMLConstants.SCHEMA_URI.equals(uri))
		{
			if (elementHit && ISMLConstants.ELEMENT_ELEMENT.equals(localName))
			{
				elementHit = false;
			}
		}
		
	}

	private void handlePhenicStartElement(String uri, String localName, Attributes attributes)
	{
		if (ISMLConstants.SMLIF_URI.equals(uri))			
		{
			if (!documentElementHit && ISMLConstants.DOCUMENT_ELEMENT.equals(localName))
			{
				documentElementHit = true;
				if (currentAliases != null)
					currentAliases.clear();
			}
			else if (documentElementHit && ISMLConstants.DATA_ELEMENT.equals(localName))
			{
				dataElementHit = true;
			}
			else if (documentElementHit && ISMLConstants.ALIAS_ELEMENT.equals(localName))
			{
				aliasElementHit = true;				
			}
		}
		else if (dataElementHit)
		{
			currentDepth++;
			if (documentNodeInx.size() - 1 < currentDepth)
			{
				currentInx = 0;
			}
			else
			{
				currentInx = ((Integer)documentNodeInx.get(currentDepth)).intValue() + 1;
			}
			
			documentNodeInx.add(new Integer(currentInx));
			if (identityStruct.isConstrained(uri, localName))
			{
				int[] primitiveIndices = new int[documentNodeInx.size()]; 
				for (int i = 0; i < primitiveIndices.length; i++)
				{
					primitiveIndices[i] = ((Integer)documentNodeInx.get(i)).intValue();
				}
				
				if (currentAliases == null || currentAliases.isEmpty())
				{
					identityStruct.addOrphanedConstrainedInstance(orphanInx++, primitiveIndices);
				}
				else
				{
					for (int i = 0, aliasCount = currentAliases.size(); i < aliasCount; i++)
					{
						identityStruct.addConstrainedInstance((String)currentAliases.get(i), primitiveIndices);
					}					
				}
			}
		}
	}

	
	private void handlePhenicCloseElement(String uri, String localName)
	{
		if (ISMLConstants.SMLIF_URI.equals(uri))
		{
			if (dataElementHit && ISMLConstants.DATA_ELEMENT.equals(localName))
			{
				dataElementHit = false;
			}
			else if (documentElementHit && ISMLConstants.DOCUMENT_ELEMENT.equals(localName))
			{
				documentElementHit = false;
				currentDepth = -1;
				documentNodeInx = new LinkedList();
			}
			else if (aliasElementHit && ISMLConstants.ALIAS_ELEMENT.equals(localName))
			{
				aliasElementHit = false;
				if (currentAlias != null)
				{
					if (currentAliases == null)
						currentAliases = new ArrayList();
					currentAliases.add(currentAlias);
				}
				currentAlias = null;
			}
		}
		else if (dataElementHit)
		{
			currentDepth--;
			if (documentNodeInx.size() - currentDepth >= 3)
				documentNodeInx.removeLast();			
		}		
	}
	
	/**
	 * @see org.xml.sax.helpers.DefaultHandler#characters(char[], int, int)
	 */
	public void characters(char[] ch, int start, int length) throws SAXException
	{
		if (aliasElementHit)
		{
			currentAlias = currentAlias == null ? new String(ch, start, length) : currentAlias + new String(ch, start, length);
		}
	}
	
	
	/**
	 * @see org.eclipse.cosmos.rm.validation.internal.databuilders.IDataBuilder#getDataStructure()
	 */
	public Object getDataStructure()
	{
		return identityStruct;
	}

	
	/**
	 * A data structure which stores information about the identity
	 * constraints.  e.g. The declaration and the instances that require
	 * applying the identity constraint to.
	 * 
	 * @author Ali Mehregani
	 */
	public static class IdentityConstraintStructure
	{		
		/**
		 * The declarations
		 * KEY = A {@link String} indicating the URI
		 * VALUE = A {@link Map}, where
		 * 	KEY = A {@link String} indicating the local name of the element
		 *  VALUE = An object of type {@link List} where each object contained is of type
		 *  {@link IdentityConstraintDecl}
		 */
		private Map declarations;
				
		/**
		 * The instances of nodes that require an identity constraint check
		 * KEY = A {@link String} indicating the document alias
		 * VALUE = A {@link List} containing elements of type int[], which identify the
		 * indices leading to the node that needs the identity constraint check. 
		 */
		private Map constrainedInstances;
		
		/**
		 * Constrained instances that don't have an alias
		 * KEY = A {@link String} indicating the index of the orphaned document
		 * VALUE = A {@link List} of elements containing objects of type int[], which identify
		 * the indices leading to the node that needs an identity constraint check.
		 */
		private Map orphanedConstrainedInstances;
		
		
		public IdentityConstraintStructure()
		{
			declarations = new Hashtable();
			constrainedInstances = new Hashtable();
			orphanedConstrainedInstances = new Hashtable();
		}
		
		/**
		 * Adds a declaration to this structure
		 * 
		 * @param uri The uri of the element
		 * @param localName The local name of the element
		 * @param identityDecl The identity declaration
		 */
		public void addDeclaration(String uri, String localName, IdentityConstraintDecl identityDecl)
		{
			Map declarationsPerURI = SMLCommonUtil.retrieveNestedMap(declarations, uri, true);
			List listOfConstraints = (List) declarationsPerURI.get(localName);
			if (listOfConstraints == null)
			{
				listOfConstraints = new ArrayList();
				declarationsPerURI.put(localName, listOfConstraints);
			}
			listOfConstraints.add(identityDecl);
		}

		
		/**
		 * Adds a constrained instance to this structure
		 * 
		 * @param alias The alias of the document
		 * @param indices The indices leading to the node that has an identity constraint
		 */
		public void addConstrainedInstance(String alias, int[] indices)
		{			
			addToNestedList (constrainedInstances, alias, indices);
		}
		
		
		public void addOrphanedConstrainedInstance(int orphanInx, int[] indices)
		{
			addToNestedList (orphanedConstrainedInstances, String.valueOf(orphanInx), indices);
		}
		
		private void addToNestedList(Map map, String key, Object value)
		{
			List nestedList = (List)map.get(key);
			if (nestedList == null)
			{
				nestedList = new ArrayList();
				map.put(key, nestedList);
			}
			nestedList.add(value);
		}
		
		/**
		 * Retunrns true iff the element with the namespace, 'uri', and local name,
		 * 'localName' has an associated identity constraint.
		 *  
		 * @param uri The URI of the element
		 * @param localName The local name of the element
		 * @return true if the element has an associated identity constraint; false
		 * otherwise
		 */
		public boolean isConstrained(String uri, String localName)
		{
			return retrieveConstraint(uri, localName) != null;			
		}
	
		
		/**
		 * Retrieves the constraints based on the uri and the local name of
		 * an element.
		 * 
		 * @param uri The uri of an element
		 * @param localName The local name of an element
		 * @return The constraints associated with the element or null if no constraints
		 * are associated.
		 */
		public IdentityConstraintDecl[] retrieveConstraint (String uri, String localName)
		{
			if (uri == null || localName == null)
				return null;
			
			Map declarationsPerURI = (Map)declarations.get(uri);
			List constraints = declarationsPerURI == null ? null : (List)declarationsPerURI.get(localName);
			return constraints == null ? null : (IdentityConstraintDecl[]) constraints.toArray(new IdentityConstraintDecl[constraints.size()]);
		}
		
		/**
		 * @return the constrainedInstances
		 */
		public Map getConstrainedInstances()
		{
			return constrainedInstances;
		}

		/**
		 * @return the declarations
		 */
		public Map getDeclarations()
		{
			return declarations;
		}

		/**
		 * @return the orphanedConstrainedInstances
		 */
		public Map getOrphanedConstrainedInstances()
		{
			return orphanedConstrainedInstances;
		}

		/**
		 * @param constrainedInstances the constrainedInstances to set
		 */
		public void setConstrainedInstances(Map constrainedInstances)
		{
			this.constrainedInstances = constrainedInstances;
		}

		/**
		 * @param declarations the declarations to set
		 */
		public void setDeclarations(Map declarations)
		{
			this.declarations = declarations;
		}

		/**
		 * @param orphanedConstrainedInstances the orphanedConstrainedInstances to set
		 */
		public void setOrphanedConstrainedInstances(Map orphanedConstrainedInstances)
		{
			this.orphanedConstrainedInstances = orphanedConstrainedInstances;
		}		
	}

	
	/**
	 * Represents an identity constraint declaration
	 * 
	 * @author Ali Mehregani
	 */
	public static class IdentityConstraintDecl
	{
		/**
		 * Indicates the key identity constraint
		 */
		public static final byte KEY_TYPE = 0x00;
		
		/**
		 * Indicates the key ref identity constraint
		 */
		public static final byte KEY_REF_TYPE = 0x01;
		
		/**
		 * Indicates the unique identity constraint
		 */
		public static final byte UNIQUE_TYPE = 0x02;
		
	
		/**
		 * The type of this declaration
		 */
		private byte type;
		
		/**
		 * The name of this declaration
		 */
		private String name;
		
		/**
		 * The fields associated with this constraint
		 */
		private List fields;
		
		/**
		 * The xpath expression indicating the selector of this constraint
		 */
		private String selector;
		
		/**
		 * The namespace context
		 */
		private NamespaceContext namespaceContext; 
		
		/**
		 * The referenced constraint
		 */
		private String reference;
		
		
		/**
		 * Constructor
		 * 
		 * @param name The name of the declaration
		 */
		public IdentityConstraintDecl(String name)
		{
			this.name = name;
			this.fields = new ArrayList();
		}

		
		public byte getType()
		{
			return type;
		}
		
		public void setType(byte type)
		{
			this.type = type;
		}

		public List getFields()
		{			
			return fields;
		}

		public String getSelector()
		{
			return selector;
		}

		public void setFields(List fields)
		{
			this.fields = fields;
		}

		public void addField (String field)
		{
			fields.add(field);
		}
		
		public void setSelector(String selector)
		{
			this.selector = selector;
		}
		
		public String getName()
		{
			return name;
		}
	
		public void setName(String name)
		{
			this.name = name;
		}

		/**
		 * @return the namespaceContext
		 */
		public NamespaceContext getNamespaceContext()
		{
			return namespaceContext;
		}

		/**
		 * @param namespaceContext the namespaceContext to set
		 */
		public void setNamespaceContext(NamespaceContext namespaceContext)
		{
			this.namespaceContext = namespaceContext;
		}

		/**
		 * @return the reference
		 */
		public String getReference()
		{
			return reference;
		}

		/**
		 * @param reference the reference to set
		 */
		public void setReference(String reference)
		{
			this.reference = reference;
		}		
		
		
	}
}
