/**********************************************************************
 * Copyright (c) 2006 IBM Corporation and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 * IBM - Initial API and implementation
 **********************************************************************/
package org.eclipse.cosmos.rm.validation.internal.reference;

import java.util.ArrayList;
import java.util.List;

import javax.xml.namespace.NamespaceContext;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.eclipse.cosmos.rm.validation.internal.common.SMLCommonUtil;
import org.eclipse.cosmos.rm.validation.internal.common.SMLMessages;
import org.eclipse.cosmos.rm.validation.internal.databuilders.NamespaceContextBuilder;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * The XPointer scheme is used to evaluate the parenthesized expressions that
 * appears after the keyword xpointer in an XPointer expression.
 * 
 * @author Ali Mehregani
 */
public class XPointerScheme extends AbstractScheme
{
	/**
	 * Represents the keyword to identify this scheme
	 */
	public static final String SCHEME = "xpointer";

	/**
	 * The xpath object used in compiling and evaluating XPath expressions
	 */
	private static XPath xpath;	

	/**
	 * @see org.eclipse.cosmos.rm.validation.internal.reference.IXScheme#evaluate(java.lang.Object)
	 */
	public Object evaluate(Object context) throws BadExpressionException
	{
		if (xpath == null)
		{
			XPathFactory factory = XPathFactory.newInstance();
			if (factory == null)
				throw new BadExpressionException(SMLMessages.errorMissingXPathFactory);
			
		    xpath = factory.newXPath();		
		    if (xpath == null)
		    	throw new BadExpressionException(SMLMessages.errorMissingXPathInstance);
		}
      
        Object namespaceContext = SMLCommonUtil.retrieveDataStructure(NamespaceContextBuilder.ID);
        if (namespaceContext instanceof NamespaceContext)
        	xpath.setNamespaceContext((NamespaceContext)namespaceContext);
        
        List listOfNodeList = new ArrayList();
        try
        {
        	String xpathExpStr = getExpression();
        	if (xpathExpStr != null)
        		xpathExpStr = xpathExpStr.replace('', '\'');
	        XPathExpression xpathExpr = xpath.compile(xpathExpStr);
	        if (xpathExpr == null)
	        	throw new BadExpressionException (SMLMessages.errorBadXPathExpression);
	        	        
	        if (context instanceof Node)
	        {
	        	evaluateXPathExpression(xpathExpr, (Node)context, listOfNodeList);       	
	        }
	        else if (context instanceof NodeList[])
	        {
	        	NodeList[] contextNodeList = (NodeList[])context;
	        	
	        	/* For every node list */
	        	for (int i = 0; i < contextNodeList.length; i++)
				{
					/* For every node in the node list */
	        		for (int j = 0, nodeListCount = contextNodeList[i].getLength(); j < nodeListCount; j++)
					{
	        			evaluateXPathExpression(xpathExpr, contextNodeList[i].item(j), listOfNodeList);       	
					}
				}
	        }
        }
        catch (XPathExpressionException xe)
        {
        	throw new BadExpressionException(xe);        	
        }
          
        return (NodeList[])listOfNodeList.toArray(new NodeList[listOfNodeList.size()]);
	}

	private void evaluateXPathExpression(XPathExpression xpathExpr, Node context, List listOfNodeList) throws XPathExpressionException
	{
    	NodeList nodeList = (NodeList)xpathExpr.evaluate(context, XPathConstants.NODESET);
    	if (nodeList != null)
    		listOfNodeList.add(nodeList); 
	}

	/**
	 * @see org.eclipse.cosmos.rm.validation.internal.reference.IXScheme#getType()
	 */
	public String getType()
	{
		return SCHEME;
	}
}
