/**********************************************************************
 * Copyright (c) 2006 IBM Corporation and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 * IBM - Initial API and implementation
 **********************************************************************/
package org.eclipse.cosmos.rm.validation.internal.smlvalidators;

import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.eclipse.cosmos.rm.validation.internal.artifacts.DocumentNode;
import org.eclipse.cosmos.rm.validation.internal.artifacts.ElementEdge;
import org.eclipse.cosmos.rm.validation.internal.artifacts.TypeNode;
import org.eclipse.cosmos.rm.validation.internal.common.IValidationOutput;
import org.eclipse.cosmos.rm.validation.internal.common.SMLCommonUtil;
import org.eclipse.cosmos.rm.validation.internal.common.SMLMessages;
import org.eclipse.cosmos.rm.validation.internal.core.AbstractSMLValidator;
import org.eclipse.cosmos.rm.validation.internal.databuilders.AcyclicDataTypesList;
import org.eclipse.cosmos.rm.validation.internal.databuilders.DataBuilderRegistry;
import org.eclipse.cosmos.rm.validation.internal.databuilders.IDataBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.ReferenceGraphDataBuilder;
import org.eclipse.cosmos.rm.validation.internal.databuilders.TypeInheritanceDataBuilderImpl;
import org.eclipse.osgi.util.NLS;


/**
 * This class is used to validate the acyclic sml extension 
 * 
 * @author Ali Mehregani
 */
public class AcyclicValidator extends AbstractSMLValidator
{
	/**
	 * Stores the node order 
	 * KEY = element type
	 * VALUE = An object of type {@link Map} that stores the following:
	 * 		KEY = Hash code of a {@link DocumentNode} representing a document node in a reference graph
	 * 		VALUE = An object of type {@link Integer} indicating the node order for a specific element type
	 */
	private Map nodeOrder;
	
	/**
	 * A map storing the tree edges that will be used to construct
	 * the cycle once a back edge is detected.
	 * 
	 * KEY = A {@link String} indicating a document alias
	 * VALUE = An object of type {@link Node}
	 */
	private Map treeForest;
	
	/**
	 * Detected cycle
	 */
	private List cycle;
	
	/**
	 * Constructor
	 */
	public AcyclicValidator()
	{
		treeForest = new Hashtable();
	}
	
	/**
	 * @see org.eclipse.cosmos.rm.validation.internal.launcher.IValidator#initialize(java.util.Map)
	 */
	public void initialize(Map validationAttribute)
	{
		super.initialize(validationAttribute);
		
		/* Register the right data builders with the registry */
		DataBuilderRegistry databuilderRegistry = DataBuilderRegistry.instance();
		databuilderRegistry.registerDataStructureBuilder(AcyclicDataTypesList.ID, new AcyclicDataTypesList());
		databuilderRegistry.registerDataStructureBuilder(ReferenceGraphDataBuilder.ID, new ReferenceGraphDataBuilder());
	}
	

	/**
	 * @see org.eclipse.cosmos.rm.validation.internal.launcher.IValidator#validate()
	 */
	public boolean validate()
	{
		/* Start the validation process for the acyclic extension */		
		Iterator acyclicTypes = (Iterator)SMLCommonUtil.retrieveDataStructure(AcyclicDataTypesList.ID);
		Map referenceGraph = (Map)SMLCommonUtil.retrieveDataStructure(ReferenceGraphDataBuilder.ID);		
		Map inheritanceMap = (Map)SMLCommonUtil.retrieveDataStructure(TypeInheritanceDataBuilderImpl.ID);
		IDataBuilder inheritanceDataBuilder = DataBuilderRegistry.instance().getDataStructureBuilder(TypeInheritanceDataBuilderImpl.ID);
		
		IValidationOutput output = getValidationOutput();		
		/* We can't continue if the reference graph is not available */
		if (referenceGraph == null || acyclicTypes == null || inheritanceMap == null)
		{
			/* Output error */
			output.reportError(ValidationMessages.acyclicUnknown);
			output.reportError(referenceGraph == null ? ValidationMessages.acyclicNoReferenceGraph : 
								acyclicTypes == null ? ValidationMessages.acyclicNoAcyclicTypeList : 
								ValidationMessages.acyclicNoInheritanceMap);
			
			return false;
		}
		
		/* Check the validity of the inheritance map */
		if (!inheritanceDataBuilder.isStructureValid())
		{
			output.reportError(inheritanceDataBuilder.getErrorMessage());
			return false;
		}
		
		String[] graphNodes = new String[referenceGraph.size()];
		int counter = 0;
		for (Iterator iterator = referenceGraph.keySet().iterator(); iterator.hasNext();)
		{			
			graphNodes[counter++] = (String)iterator.next();			
		}
		
		/* Iterate through the acyclic element types */
		Edge backEdge = null;
		cycle = null;
		for (Iterator iter = acyclicTypes; iter.hasNext() && backEdge == null;)
		{
			TypeNode elementType = (TypeNode)iter.next();
			treeForest.clear();

			
			/* Because there is no guarantee that the graph is connected, we'll need to iterate
			 * through each node. */
			for (int i = 0; i < graphNodes.length && backEdge == null; i++)
			{
				/* Reset the order for the element type */
				if (nodeOrder != null)
					nodeOrder.remove(elementType);
				
				if (getNodeOrder(referenceGraph, graphNodes[i], elementType) == -1)
				{
					try
					{
						backEdge = dfsSearch (referenceGraph, elementType, graphNodes[i]);
					} 
					catch (URISyntaxException e)
					{
						output.reportError(SMLMessages.errorInvalidURI);
						if (e.getMessage() != null)
							output.reportError(e.getMessage());
						return false;
					}
				}
			}			
		}
		
		/* If a back edge was detected, then the reference graph is not acyclic */
		if (backEdge != null)
		{
			output.reportError(NLS.bind(ValidationMessages.acyclicCycleFound, backEdge.referenceType));
			if (cycle != null && cycle.size() > 0)
			{
				String cycleString = "";
				for (int i = 0, cycleNodeCount = cycle.size(); i < cycleNodeCount; i++)
				{
					cycleString += cycle.get(i) + ", ";
				}
				cycleString = cycleString.substring(0, cycleString.length() - 2);
				output.reportError(NLS.bind(ValidationMessages.acyclicDetectedCycle, cycleString));
			}
			return false;
		}
		
		return true;
	}

	
	/**
	 * Return the node order of the node that is passed in.
	 * 
	 * @param node The node order
	 * @param elementType The element name indicating the type that the reference
	 * graph should be based on
	 *  
	 * @return An integer indicating the node order
	 */
	private int getNodeOrder(Map referenceGraph, String node, TypeNode elementType)
	{		
		DocumentNode documentNode = (DocumentNode)referenceGraph.get(node);
		if (nodeOrder == null || documentNode == null)
			return -1;
				
		Map ordersPerElement = (Map)nodeOrder.get(elementType);
		Integer order = ordersPerElement == null ? null : (Integer)ordersPerElement.get(new Integer(documentNode.hashCode()));
		return order == null ? -1 : order.intValue();
	}

	
	/**
	 * Increment the node order of the 'node' passed in.
	 * 
	 * @param referenceGraph The reference graph
	 * @param elementType The element name represeting the type of the reference
	 * @param node  The node
	 */
	private void incrementNodeOrder (Map referenceGraph, TypeNode elementType, String node)
	{
		DocumentNode documentNode = (DocumentNode)referenceGraph.get(node);
		if (documentNode == null)
			return;
		
		if (nodeOrder == null)
			nodeOrder = new Hashtable();
		
		Integer hashcode = new Integer(documentNode.hashCode());
		Map ordersPerElement = (Map)nodeOrder.get(elementType);
		if (ordersPerElement == null)
		{
			ordersPerElement = new Hashtable();
			nodeOrder.put(elementType, ordersPerElement);
		}
				
		Integer order = (Integer)ordersPerElement.get(hashcode);
		ordersPerElement.put(hashcode, order == null ? new Integer(0) : new Integer(order.intValue() + 1));
	}

	
	/**
	 * Perform a DFS walk through of the reference graph to determine if there
	 * are any back edges.  
	 * 
	 * @param referenceGraph The reference graph
	 * @param elementType The element name
	 * @param documentAlias The node
	 * @return The first back edge detected
	 * @throws URISyntaxException 
	 */
	private Edge dfsSearch(Map referenceGraph, TypeNode elementType, String documentAlias) throws URISyntaxException
	{
		DocumentNode documentNode = (DocumentNode)referenceGraph.get(documentAlias);
		if (documentNode == null)
			return null;
		
		incrementNodeOrder (referenceGraph, elementType, documentAlias);
		ElementEdge[] edges = documentNode.getEdges(elementType.getUri(), elementType.getType());
		Edge backEdge = null;
		
		Node node = (Node)treeForest.get(documentAlias);
		if (node == null)
		{
			node = new Node();
			treeForest.put(documentAlias, node);
		}
		
		for (int i = 0; i < edges.length; i++)
		{
			String document = edges[i].getDocument();
						
			if (getNodeOrder(referenceGraph, document, elementType) == -1)
			{	
				node.addEdge(new Edge(documentAlias, document, elementType.getUri()+elementType.getType()));
				backEdge = dfsSearch (referenceGraph, elementType, document);				
			}
			else
			{
				backEdge = new Edge(documentAlias, document, elementType.getUri()+elementType.getType());
				
				/* A back edge was detected, find the cycle */
				List cycle = new ArrayList();
				cycle.add(documentAlias);
				cycle.add(document);
				findCycle(document, documentAlias, cycle);
				
				if (documentAlias.equals(cycle.get(cycle.size() - 1)))
					this.cycle = cycle;
				
				return backEdge;
			}
		}
		
		return backEdge;
	}
	
	
	private void findCycle(String source, String target, List cycle)
	{
		Node node = (Node)treeForest.get(source);
		if (node == null)
			return;
		
		for (int i = 0, edgeCount = node.edges.size(); i < edgeCount; i++)
		{			
			Edge currentEdge = (Edge)node.edges.get(i);
			cycle.add(currentEdge.targetDocument);
			if (target.equals(currentEdge.targetDocument))
				break;
			
			findCycle(currentEdge.targetDocument, target, cycle);
			if (cycle.size() > 0 && target.equals(cycle.get(cycle.size() - 1)))
				break;
			
			cycle.remove(currentEdge.targetDocument);
		}
	}


	private static class Edge
	{
		public String sourceDocument;
		public String targetDocument;
		public String referenceType;
		
		public Edge(String sourceDocument, String targetDocument, String referenceType)
		{
			this.sourceDocument = sourceDocument;
			this.targetDocument = targetDocument;
			this.referenceType = referenceType;		
		}
	}
	
	
	private static class Node
	{
		public List edges;
		
		public Node()
		{
			edges = new ArrayList();
		}
		
		public void addEdge (Edge edge)
		{
			edges.add(edge);
		}
	}
}
